import { BC_TERRAIN_SETTINGS } from "../GlobalVariables/GlobalVariables";
import { ChunkStorageTypes } from "../WorldGeneration/WorldChunk/WorldGenChunk";
import { getTileAt } from "../WorldGeneration/WorldGen";
import { PointInt2D } from "./Math.utils";
import { SceneObject } from "../SceneObjects/SceneObject";

class PathFinderNode {
    position;
    fScore = 0;
    gScore = 1e16;
    dScore = 1;
    // hScore = 1e16;
    id;

    /**
     *
     * @param {PointInt2D} position
     * @param {Number} gScore
     */
    constructor(position, gScore = 1e16, fScore = 1e16, dScore) {
        this.position = position;
        this.id = position.getX() + "_" + position.getY();
        this.gScore = gScore;
        this.fScore = fScore;
        this.dScore = dScore;
    }
}

export class NavigationPath {
    path;

    /**
     * @param {Array<PointInt2D>} path 
     */
    constructor(path = [])
    {
        this.path = path;
    }
};

export class PathFinder {
    /**
     * @type Array<PathFinderNode>
     */
    _openSet = new Array();
    _closedSet = new Array();

    _goal = new PointInt2D();
    _start = new PointInt2D();

    /**
     *
     * @param {PointInt2D} start
     * @param {PointInt2D} goal
     */
    search(start, goal) {
        start.divideBy(BC_TERRAIN_SETTINGS.totalSize);
        start.multiplyBy(BC_TERRAIN_SETTINGS.totalSize);
        this._start = start;

        goal.divideBy(BC_TERRAIN_SETTINGS.totalSize);
        goal.multiplyBy(BC_TERRAIN_SETTINGS.totalSize);
        this._goal = goal;

        this._openSet.push(new PathFinderNode(this._start, 0, this._heuristic(this._start, 0)));

        //code
        /**
         * @type PathFinderNode
         */
        let currentNode;
        // /**
        //  * @type Map<String, Number>
        //  */
        // let gScores = new Map();
        // /**
        //  * @type Map<String, Number>
        //  */
        // let fScores = new Map();

        /**
         * @type Map<String, PathFinderNode>
         */
        let cameFrom = new Map();

        let minFScoreNodeIndex = 0;
        //START LOOP WOOOOAAAW
        // console.log(goal);
        while (this._openSet.length > 0) {
            //find node with min f score
            //better to rewrite it to priority queue
            for (let i = 0; i < this._openSet.length; i++) {
                if(!this._openSet[i]){
                    return undefined;
                }
                if (this._openSet[i].fScore < this._openSet[minFScoreNodeIndex].fScore) minFScoreNodeIndex = i;
            }
            //wow! node is found and set!!
            currentNode = this._openSet[minFScoreNodeIndex];

            if (PointInt2D.isEqual(currentNode.position, this._goal)) {
                //wow!!! we have found an end of the path!!! this is so cool!!!
                // console.log(cameFrom);
                return new NavigationPath(this._reconstructPath(cameFrom, currentNode)); //return something weird stuff
            }

            //and now we must delete this node... what a sad situation...
            this._openSet.splice(minFScoreNodeIndex, 1);
            //but wait! we add it to closed set! nice!!! (why we are doing this??? idk... okay)
            this._closedSet.push(currentNode);

            /**
             * @type Array<PathFinderNode>
             */
            let currentNeighbors = this._getNeighbors(currentNode.position);
            if(!currentNeighbors) return undefined;
            // console.log(currentNeighbors);
            for (const neighbor of currentNeighbors) {
                this._closedSet.push(neighbor);
                let tentativeGScore = currentNode.gScore + this._manhattanDistance(currentNode.position, neighbor.position);
                if (tentativeGScore < neighbor.gScore) {
                    // console.log(neighbor.id, currentNode.id);
                    cameFrom.set(neighbor.id, currentNode);
                    neighbor.gScore = tentativeGScore;
                    neighbor.fScore = tentativeGScore + this._heuristic(neighbor.position, neighbor.dScore);

                    let neighborFoundInOpenSet = false;
                    for (const node of this._openSet) {
                        if (node.id === neighbor.id) {
                            neighborFoundInOpenSet = true;
                            break;
                        }
                    }
                    if (!neighborFoundInOpenSet) this._openSet.push(neighbor);
                }
            }
        }

        return undefined;
    }

    /**
     *
     * @param {Map<String, PathFinderNode>} cameFrom
     * @param {PathFinderNode} current
     */
    _reconstructPath(cameFrom, current) {
        // console.log(cameFrom);
        let totalPath = [current.position];
        let keys = [...cameFrom.keys()];
        while (keys.includes(current.id)) {
            current = cameFrom.get(current.id);
            totalPath.push(current.position);
        }
        return totalPath.reverse();
    }

    /**
     *
     * @param {PointInt2D} root
     */
    _getNeighbors(root) {
        /**
         * @type Array<PathFinderNode>
         */
        let neighbors = new Array();
        /**
         * @type SceneObject
         */
        let currentTile = getTileAt(root.getX(), root.getY() + BC_TERRAIN_SETTINGS.totalSize, ChunkStorageTypes.TYPE_TERRAIN);
        if(!currentTile) return undefined;
        // currentTile.drawObject.tint = 0xff0000;
        let node = new PathFinderNode(new PointInt2D(currentTile.worldPosition.getX(), currentTile.worldPosition.getY()), 1e16, 1e16, currentTile.props.navigationCost);
        //north
        if (!this._existsInClosedSet(node)) {
            neighbors.push(node);
        }
        //south
        currentTile = getTileAt(root.getX(), root.getY() - BC_TERRAIN_SETTINGS.totalSize, ChunkStorageTypes.TYPE_TERRAIN);
        if(!currentTile) return undefined;
        // currentTile.drawObject.tint = 0xff0000;
        node = new PathFinderNode(new PointInt2D(currentTile.worldPosition.getX(), currentTile.worldPosition.getY()), 1e16, 1e16, currentTile.props.navigationCost);
        if (!this._existsInClosedSet(node)) {
            neighbors.push(node);
        }
        //east
        currentTile = getTileAt(root.getX() + BC_TERRAIN_SETTINGS.totalSize, root.getY(), ChunkStorageTypes.TYPE_TERRAIN);
        if(!currentTile) return undefined;
        // currentTile.drawObject.tint = 0xff0000;
        node = new PathFinderNode(new PointInt2D(currentTile.worldPosition.getX(), currentTile.worldPosition.getY()), 1e16, 1e16, currentTile.props.navigationCost);
        if (!this._existsInClosedSet(node)) {
            neighbors.push(node);
        }
        //west
        currentTile = getTileAt(root.getX() - BC_TERRAIN_SETTINGS.totalSize, root.getY(), ChunkStorageTypes.TYPE_TERRAIN);
        if(!currentTile) return undefined;
        // currentTile.drawObject.tint = 0xff0000;
        node = new PathFinderNode(new PointInt2D(currentTile.worldPosition.getX(), currentTile.worldPosition.getY()), 1e16, 1e16, currentTile.props.navigationCost);
        if (!this._existsInClosedSet(node)) {
            neighbors.push(node);
        }

        return neighbors;
    }

    /**
     *
     * @param {PointInt2D} start
     * @param {PointInt2D} goal
     * @returns
     */
    _manhattanDistance(start, goal) {
        return Math.abs(start.getX() - goal.getX()) + Math.abs(start.getY() - goal.getY());
    }

    /**
     *
     * @param {PointInt2D} current
     * @param {Number} d
     */
    _heuristic(current, d) {
        let x1 = current.getX() - this._goal.getX();
        let y1 = current.getY() - this._goal.getY();
        let x2 = this._start.getX() - this._goal.getX();
        let y2 = this._start.getY() - this._goal.getY();
        let cross = Math.abs(x1*y2 - x2*y1);
        return this._manhattanDistance(current, this._goal) * d + cross*0.001;
    }

    /**
     *
     * @param {PathFinderNode} node
     */
    _existsInClosedSet(node) {
        for (const i of this._closedSet) {
            if (i.id === node.id) return true;
        }
        return false;
    }
}
